package othello.observation;

import java.util.List;
import java.util.ArrayList;

/** A class that informs observers potentially asynchronously.
 *  Each observer is sent an object of class CountMonitorInterface
 *  and is expected to call "decrement" on that method exactly once.
 *  Once all the observers have called decrement, the notifying object
 *  (the object of this class) proceeds.
 * 
 * <p>Title: Othello</p>
 * <p>Description: </p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Memorial University</p>
 * @author Theodore S. Norvell
 * @version 1.0
 */

public class ConcurrentlyObservable {
	
	private ReadersAndWriters rw = new ReadersAndWriters() ;

    private List<ConcurrentObserver> observerList
        = new ArrayList<ConcurrentObserver>() ;
    
    protected void readerEnters () { rw.readerEnters() ; }
    
    protected void readerExits() { rw.readerExits() ; }
    
    protected void writerEnters() { rw.writerEnters() ; }
    
    protected void writerExits() { rw.writerExits() ; }
    
    protected void writerGoesPassive() { rw.writerGoesPassive() ; }
    
    protected void writerGoesActive() { rw.writerGoesActive() ; }
    
    public void addObserver( ConcurrentObserver observer ) {
    	writerEnters() ;
    	try {
    		System.err.println("Adding observer "+observer);
    		observerList.add( observer ); }
    	finally{ writerExits() ; } }
        
    /** Notify all observers of a change.
     * This routine should only be called by a writer.
     * Observers may read, but must not write.
     * @param arg -- Passed on to the observer.
     */
    protected  void notifyObservers( Object arg ) {
    	writerGoesPassive() ;
        System.err.println("Notifying observers");
        CountMonitor countMonitor = new CountMonitor( observerList.size() ) ;
        for(ConcurrentObserver observer : observerList ) {
            observer.update( countMonitor, arg ) ; }
        countMonitor.waitForZero() ; 
		writerGoesActive() ; }
}

class CountMonitor implements CountMonitorInterface {
    private int count ;
    
    CountMonitor( int start ) { count = start ; }
    
    public synchronized void decrement() {
        count -= 1 ;
        if( count == 0 ) notifyAll() ; }
    
    synchronized void waitForZero() {
        if( count > 0 ) 
            try { wait() ; }
            catch( InterruptedException e ) { } }
}
