package othello.view;

import java.util.Observable ;

import othello.model.GameModelInterface;
import othello.model.Pair;
import othello.model.GameConstants;
import othello.observation.* ;

/** The animator object caches the state of the board
 *  and animates transitions between the 
 * <p>Title: Othello</p>
 * <p>Description: </p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Memorial University</p>
 * @author Theodore S. Norvell
 * @version 1.0
 */

public class Animator extends Observable
implements ConcurrentObserver, GameConstants
{    
	private int [][] square = new int[ROWS][COLS] ;
	{
		for( int r=0 ; r<ROWS ; ++r ) {
			for( int c=0 ; c<COLS ; ++c ) {
				square[r][c] = EMPTY ; } }
	}

	public final int QUIET_STATE = 0 ;
	public final int FLYING_STATE = 1 ;
	public final int FLIPPING_STATE = 2 ;

	private int state = QUIET_STATE ;
	private int flippingRow ;
	private int flippingCol ;
	private double amountOfFlip ;
	private double flyingRow ;
	private double flyingCol ;
	private int newColour ;

	private GameModelInterface gameModel ;

	public Animator( GameModelInterface gameModel ) {
		this.gameModel = gameModel ;
		gameModel.addObserver(this);}

	public void update( CountMonitorInterface countMonitor, Object arg ) {
		Pair pair = (Pair) arg ;
		int row = pair.getRow() ;
		int col = pair.getCol() ;
		int oldColour ; synchronized( Animator.this ) {
			oldColour = square[ row ][ col ] ; }
		int newColour = gameModel.getPieceAt( row, col ) ;
		if( oldColour == EMPTY ) {
			// This is a new piece being laid.
			animateFlyIn( row, col, newColour, countMonitor ) ;
		}
		else {
			// This is a flip.
			animateFlip( row, col, newColour, countMonitor ) ;
		}

	}

	synchronized public int getPieceAt( int r, int c ) { return square[r][c] ; }

	synchronized public int getState() { return state ; }

	/** Flipping amount varies from 0 (not at all) to 1 (fully flipped) */
	synchronized public double getFlippingAmount() { return amountOfFlip ; }

	synchronized public int getFlippingRow() { return flippingRow ; }

	synchronized public int getFlippingCol() { return flippingCol ; }

	synchronized public double getFlyingRow() { return flyingRow ; }

	synchronized public double getFlyingCol() { return flyingCol ; }

	synchronized public int getNewColour() { return newColour ; }

	private void animateFlyIn( final int row, final int col,
			final int newColour,
			final CountMonitorInterface counter ) {
		class FlyThread extends Thread {
			@Override public void run() {
				int iterations ;
				synchronized( Animator.this ) {
					state = FLYING_STATE ;
					Animator.this.newColour = newColour ;
					if( newColour == WHITE ) {
						flyingRow = -2 ;
						flyingCol = COLS / 2.0 ;
						iterations = 10*(2+row) ; }
					else {
						flyingRow = ROWS + 1 ;
						flyingCol = COLS / 2.0 ;
						iterations = 10*(2+9-row) ; } }

				double dr = (row - flyingRow) / iterations ;
				double dc = (col - flyingCol) / iterations ;

				for( int i=0 ; i<iterations ; ++i ) {
					synchronized( Animator.this ) {
						flyingRow += dr ;
						flyingCol += dc ; 
						setChanged() ; }
					notifyObservers();
					try{ sleep(20) ; }
					catch( InterruptedException e ) {}
				}

				synchronized( Animator.this ) {
					state = QUIET_STATE ;
					square[ row ][ col ] = newColour ;
					setChanged() ; }
				notifyObservers(); 
				counter.decrement(); }
		}
		Thread t = new FlyThread() ;
		t.start();
	}

	private void animateFlip( final int row, final int col,
			final int newColour,
			final CountMonitorInterface counter ) {
		class FlipThread extends Thread {
			@Override public void run() {
				synchronized( Animator.this ) {
					state = FLIPPING_STATE ;
					Animator.this.newColour = newColour ;
					square[ row ][ col ] = EMPTY ;
					flippingRow = row ;
					flippingCol = col ;
					amountOfFlip = 0.0 ; }

				int iterations = 50 ;
				double delta =  1 / (double)iterations ;
				for( int i=0 ; i<iterations ; ++i ) {
					synchronized( Animator.this ) {
						amountOfFlip += delta ; 
						setChanged() ; }
					notifyObservers();
					try{ sleep(20) ; }
					catch( InterruptedException e ) {}
				}

				synchronized( Animator.this ) {
					state = QUIET_STATE ;
					square[ row ][ col ] = newColour ;
					setChanged() ; }
				notifyObservers();
				counter.decrement(); }
		}
		Thread t = new FlipThread() ;
		t.start();
	}
}
